from ecs_deplojo import utils


def test_find_missing_services(cluster, connection):
    missing = utils.find_missing_services(
        ecs=connection.ecs,
        cluster=cluster["cluster"]["clusterName"],
        services={"service-1", "service-2"},
    )
    assert missing == {"service-1", "service-2"}


def test_find_missing_services_existing(cluster, connection, definition):
    retval = connection.ecs.register_task_definition(**definition.as_dict())
    task_definition_arn = retval["taskDefinition"]["taskDefinitionArn"]

    connection.ecs.create_service(
        cluster=cluster["cluster"]["clusterName"],
        serviceName="service-2",
        taskDefinition=task_definition_arn,
    )

    missing = utils.find_missing_services(
        ecs=connection.ecs,
        cluster=cluster["cluster"]["clusterName"],
        services={"service-1", "service-2"},
    )
    assert missing == {"service-1"}


def test_find_missing_services_paginate(cluster, connection, definition):
    retval = connection.ecs.register_task_definition(**definition.as_dict())
    task_definition_arn = retval["taskDefinition"]["taskDefinitionArn"]

    connection.ecs.create_service(
        cluster=cluster["cluster"]["clusterName"],
        serviceName="service-2",
        taskDefinition=task_definition_arn,
    )
    connection.ecs.create_service(
        cluster=cluster["cluster"]["clusterName"],
        serviceName="service-40",
        taskDefinition=task_definition_arn,
    )

    all_services = {"service-%d" % i for i in range(39)}
    missing = utils.find_missing_services(
        ecs=connection.ecs,
        cluster=cluster["cluster"]["clusterName"],
        services=all_services
    )
    all_services.remove("service-2")
    assert missing == all_services



import boto3


class Connection:
    def __init__(self, role_arn=None):
        credentials = {}
        if role_arn:
            sts = boto3.client("sts")
            resp = sts.assume_role(RoleArn=role_arn, RoleSessionName="ecs-deplojo")
            credentials.update(
                {
                    "aws_secret_access_key": resp["Credentials"]["SecretAccessKey"],
                    "aws_access_key_id": resp["Credentials"]["AccessKeyId"],
                    "aws_session_token": resp["Credentials"]["SessionToken"],
                }
            )
        self.ecs = boto3.client("ecs", **credentials)



import os

from ecs_deplojo import task_definitions

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def test_generate_task_definition(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join("task_definition.json")
    filename.write(task_data)

    task_definition = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        template_vars={"image": "my-docker-image:1.0"},
        overrides={},
        name="my-task-def",
    )
    expected = task_definitions.TaskDefinition(
        {
            "family": "my-task-def",
            "volumes": [],
            "containerDefinitions": [
                {
                    "name": "default",
                    "image": "my-docker-image:1.0",
                    "essential": True,
                    "command": ["hello", "world"],
                    "hostname": "my-task-def",
                    "memory": 256,
                    "cpu": 0,
                    "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                    "environment": {},
                }
            ],
            "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
        }
    )
    assert task_definition == expected


def test_generate_task_definition_overrides(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join("task_definition.json")
    filename.write(task_data)

    task_definition = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        template_vars={"image": "my-docker-image:1.0"},
        overrides={
            "default": {
                "memory": 512,
                "memoryReservation": 128,
                "portMappings": [{"hostPort": 80, "containerPort": 9000}],
            }
        },
        name="my-task-def",
    )
    expected = task_definitions.TaskDefinition(
        {
            "family": "my-task-def",
            "volumes": [],
            "containerDefinitions": [
                {
                    "name": "default",
                    "image": "my-docker-image:1.0",
                    "essential": True,
                    "command": ["hello", "world"],
                    "hostname": "my-task-def",
                    "memory": 512,
                    "memoryReservation": 128,
                    "cpu": 0,
                    "portMappings": [
                        {"containerPort": 8080, "hostPort": 0},
                        {"containerPort": 9000, "hostPort": 80},
                    ],
                    "environment": {},
                }
            ],
            "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
        }
    )
    assert task_definition == expected


def test_generate_multiple_task_definitions(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "web-1",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        },
        {
          "name": "web-2",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join("task_definition.json")
    filename.write(task_data)

    config = {
        "environment": {"DATABASE_URL": "postgresql://"},
        "environment_groups": {
            "group-1": {"ENV_CODE": "group-1"},
            "group-2": {"ENV_CODE": "group-2"},
        },
        "task_definitions": {
            "task-def-1": {
                "template": filename.strpath,
                "environment_group": "group-1",
                "overrides": {"web-1": {"memory": 512}},
            },
            "task-def-2": {
                "template": filename.strpath,
                "environment_group": "group-2",
                "overrides": {"web-1": {"memory": 512}},
            },
        },
    }

    result = task_definitions.generate_task_definitions(
        config, template_vars={"image": "my-docker-image:1.0"}, base_path=None
    )

    expected = {
        "task-def-1": task_definitions.TaskDefinition(
            {
                "family": "task-def-1",
                "volumes": [],
                "containerDefinitions": [
                    {
                        "name": "web-1",
                        "image": "my-docker-image:1.0",
                        "essential": True,
                        "command": ["hello", "world"],
                        "hostname": "task-def-1-web-1",
                        "memory": 512,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {
                            "DATABASE_URL": "postgresql://",
                            "ENV_CODE": "group-1",
                        },
                    },
                    {
                        "name": "web-2",
                        "image": "my-docker-image:1.0",
                        "essential": True,
                        "command": ["hello", "world"],
                        "hostname": "task-def-1-web-2",
                        "memory": 256,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {
                            "DATABASE_URL": "postgresql://",
                            "ENV_CODE": "group-1",
                        },
                    },
                ],
                "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
            }
        ),
        "task-def-2": task_definitions.TaskDefinition(
            {
                "family": "task-def-2",
                "volumes": [],
                "containerDefinitions": [
                    {
                        "name": "web-1",
                        "image": "my-docker-image:1.0",
                        "essential": True,
                        "hostname": "task-def-2-web-1",
                        "command": ["hello", "world"],
                        "memory": 512,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {
                            "DATABASE_URL": "postgresql://",
                            "ENV_CODE": "group-2",
                        },
                    },
                    {
                        "name": "web-2",
                        "image": "my-docker-image:1.0",
                        "hostname": "task-def-2-web-2",
                        "essential": True,
                        "command": ["hello", "world"],
                        "memory": 256,
                        "cpu": 0,
                        "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                        "environment": {
                            "DATABASE_URL": "postgresql://",
                            "ENV_CODE": "group-2",
                        },
                    },
                ],
                "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
            }
        ),
    }

    assert result == expected


def test_generate_task_definitions_write_output(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "web-1",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "logConfiguration": {},
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()
    base_path = tmpdir.join("input").mkdir()

    filename = base_path.join("task_definition.json")
    filename.write(task_data)

    config = {
        "environment": {"DATABASE_URL": "postgresql://"},
        "task_definitions": {
            "task-def-1": {
                "template": "task_definition.json",
                "overrides": {
                    "web-1": {
                        "memory": 512,
                        "logConfiguration": {
                            "logDriver": "awslogs",
                            "options": {
                                "awslogs-group": "default",
                                "awslogs-region": "eu-west-1",
                            },
                        },
                    }
                },
            }
        },
    }

    task_definitions.generate_task_definitions(
        config,
        template_vars={"image": "my-docker-image:1.0"},
        base_path=base_path.strpath,
        output_path=tmpdir.join("output").mkdir().strpath,
    )

    assert tmpdir.join("output").join("task-def-1.json").exists()


def test_generate_task_definition_with_task_role_arn(tmpdir):
    task_data = """
    {
      "family": "default",
      "volumes": [],
      "containerDefinitions": [
        {
          "name": "default",
          "image": "${image}",
          "essential": true,
          "command": ["hello", "world"],
          "memory": 256,
          "cpu": 0,
          "portMappings": [
            {
              "containerPort": 8080,
              "hostPort": 0
            }
          ]
        }
      ]
    }
    """.strip()

    filename = tmpdir.join("task_definition.json")
    filename.write(task_data)

    result = task_definitions.generate_task_definition(
        filename.strpath,
        environment={},
        task_role_arn="arn:my-task-role",
        template_vars={"image": "my-docker-image:1.0"},
        overrides={},
        name="my-task-def",
    )

    expected = task_definitions.TaskDefinition(
        {
            "family": "my-task-def",
            "taskRoleArn": "arn:my-task-role",
            "volumes": [],
            "containerDefinitions": [
                {
                    "name": "default",
                    "image": "my-docker-image:1.0",
                    "essential": True,
                    "hostname": "my-task-def",
                    "command": ["hello", "world"],
                    "memory": 256,
                    "cpu": 0,
                    "portMappings": [{"containerPort": 8080, "hostPort": 0}],
                    "environment": {},
                }
            ],
            "tags": [{"key": "createdBy", "value": "ecs-deplojo"}],
        }
    )
    assert result == expected



import typing


def find_missing_services(
    ecs, cluster: str, services: typing.Set[str]
) -> typing.Set[str]:
    """Return a set of service names which don't exist in AWS.

    We use `ECS.Client.describe_services` since we have a list of service
    names we want to check and instead of just retrieving all services in the
    cluster we pass the items we want. We can only pass 10 services per call
    so we iterate over the list in chunks.
    """
    existing_services = set()
    for service in describe_services(ecs, cluster, services):
        existing_services.add(service["serviceName"])
    return set(services) - existing_services


def describe_services(
    ecs, cluster: str, services: typing.Set[str]
) -> typing.List[typing.Dict[str, typing.Any]]:
    """Wrap `ECS.Client.describe_services` to allow more then 10 services in
    one call.

    """
    result: typing.List[typing.Dict[str, typing.Any]] = []
    services_list = list(services)
    for i in range(0, len(services_list), 10):
        response = ecs.describe_services(
            cluster=cluster, services=services_list[i : i + 10]
        )
        result.extend(response["services"])
    return result



import re

from setuptools import setup, find_packages


install_requires = [
    'boto3==1.9.196',
    'click==7.0',
    'pyaml==16.12.2',
    'pytz',
]

tests_requires = [
    'flake8==3.7.8',
    'isort==4.3.21',
    'moto==1.3.13',
    'pytest==5.0.1',
    'pytest-cov==2.7.1',
]

with open('README.rst') as fh:
    long_description = re.sub(
        '^.. start-no-pypi.*^.. end-no-pypi', '', fh.read(), flags=re.M | re.S)

setup(
    name='ecs-deplojo',
    version='0.8.1',
    author='Lab Digital B.V.',
    author_email='opensource@labdigital.nl',
    url='https://www.github.com/labd/ecs-deplojo/',
    description="Deployment tool for Amazon ECS",
    long_description=long_description,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=tests_requires,
    extras_require={'test': tests_requires},
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    entry_points={
        'console_scripts': {
            'ecs-deplojo = ecs_deplojo.cli:main'
        }
    },
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



import copy
import json
import operator
import os.path
import typing
from string import Template


class TaskDefinition:
    """A TaskDefinition exists out of a set of containers."""

    def __init__(self, data):
        self._data = data

    @classmethod
    def load(cls, fh) -> "TaskDefinition":
        data = json.load(fh)
        return cls(data)

    def as_dict(self):
        """Output the TaskDefinition in a boto3 compatible format.

        See the boto3 documentation on `ECS.Client.register_task_definition`.
        """
        result = copy.deepcopy(self._data)
        for container in result["containerDefinitions"]:
            container["environment"] = sorted(
                [
                    {"name": k, "value": str(v)}
                    for k, v in container.get("environment", {}).items()
                ],
                key=operator.itemgetter("name"),
            )
        return result

    def apply_variables(self, variables: typing.Dict[str, str]):
        """Interpolate all the variables used in the task definition"""
        for container in self.container_definitions:
            container["image"] = Template(container["image"]).substitute(variables)

    def apply_overrides(self, overrides):
        """Apply overrides for all containers within this task definition."""
        for container in self.container_definitions:
            container_overrides = overrides.get(container["name"], {})
            for key, value in container_overrides.items():
                if key in container and isinstance(container[key], list):
                    container[key].extend(value)
                elif key in container and isinstance(container[key], dict):
                    container[key].update(value)
                else:
                    container[key] = value

    def set_environment(self, env: typing.Dict[str, str]):
        """Interpolate all the variables used in the task definition"""
        for container in self.container_definitions:
            container["environment"] = env

    def __str__(self):
        if self._data.get("name"):
            return self._data.get("name")
        return "(unregistered)"

    def __eq__(self, other: object):
        if not isinstance(other, TaskDefinition):
            return False
        return self._data == other._data

    def __repr__(self):
        return json.dumps(self._data)

    @property
    def tags(self) -> typing.List[typing.Dict[str, str]]:
        return self._data.get("tags")

    @tags.setter
    def tags(self, value: typing.List[typing.Dict[str, str]]):
        self._data["tags"] = value

    @property
    def family(self) -> str:
        return self._data.get("family")

    @family.setter
    def family(self, value: str):
        self._data["family"] = value

    @property
    def revision(self) -> int:
        return self._data.get("revision")

    @revision.setter
    def revision(self, value: int):
        self._data["revision"] = value

    @property
    def name(self) -> str:
        return self._data.get("name")

    @name.setter
    def name(self, value: str):
        self._data["name"] = value

    @property
    def task_role_arn(self) -> str:
        return self._data.get("taskRoleArn")

    @task_role_arn.setter
    def task_role_arn(self, value: str):
        self._data["taskRoleArn"] = value

    @property
    def arn(self) -> str:
        return self._data.get("arn")

    @arn.setter
    def arn(self, value: str):
        self._data["arn"] = value

    @property
    def container_definitions(self):
        return self._data.get("containerDefinitions")

    @container_definitions.setter
    def container_definitions(self, value):
        self._data["containerDefinitions"] = value


def generate_task_definitions(
    config, template_vars, base_path, output_path=None
) -> typing.Dict[str, TaskDefinition]:
    """Generate the task definitions

    :parameter config: The yaml config contents
    :parameter template_vars: Key-Value dict with template replacements
    :parameter base_path: The base path (location of the config file)
    :parameter output_path: Optional path to write the task definitions to.
    :rtype dict:

    """
    task_definitions = {}

    for name, info in config["task_definitions"].items():
        # Create a copy of the environment dict so that it can safely be
        # modified.
        env_vars = copy.deepcopy(config.get("environment", {}))

        # Environment groups
        env_group = info.get("environment_group")
        if env_group:
            env_vars.update(config["environment_groups"][env_group])

        overrides = info.get("overrides", {})
        definition = generate_task_definition(
            filename=info["template"],
            environment=env_vars,
            template_vars=template_vars,
            overrides=overrides,
            name=name,
            base_path=base_path,
            task_role_arn=info.get("task_role_arn"),
        )

        if output_path:
            write_task_definition(name, definition, output_path)
        task_definitions[name] = definition
    return task_definitions


def generate_task_definition(
    filename: str,
    environment: typing.Dict[str, str],
    template_vars,
    overrides,
    name,
    base_path=None,
    task_role_arn=None,
) -> TaskDefinition:

    """Generate the task definitions"""
    if base_path:
        filename = os.path.join(base_path, filename)

    with open(filename, "r") as fh:
        task_definition = TaskDefinition.load(fh)

    task_definition.family = name
    if task_role_arn:
        task_definition.task_role_arn = task_role_arn

    # If no hostname is specified for the container we set it ourselves to
    # `{family}-{container-name}-{num}`
    num_containers = len(task_definition.container_definitions)
    for container in task_definition.container_definitions:
        hostname = task_definition.family
        if num_containers > 1:
            hostname += "-%s" % container["name"].replace("_", "-")
        container.setdefault("hostname", hostname)

    task_definition.set_environment(environment)
    task_definition.apply_variables(template_vars)
    task_definition.apply_overrides(overrides)
    task_definition.tags = [{"key": "createdBy", "value": "ecs-deplojo"}]

    return task_definition


def write_task_definition(name: str, definition: TaskDefinition, output_path) -> None:
    filename = os.path.join(output_path, "%s.json" % name)
    with open(filename, "w") as fh:
        json.dump(definition.as_dict(), fh, indent=4)
